/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.messaginghub.pooled.jms;

import static org.junit.jupiter.api.Assertions.assertTrue;

import javax.jms.Connection;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.MessageConsumer;
import javax.jms.Session;
import javax.jms.TemporaryQueue;

import org.apache.activemq.broker.region.RegionBroker;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInfo;
import org.junit.jupiter.api.Timeout;

/**
 * Test of lingering temporary destinations on pooled connections when the
 * underlying connections are reused. Also tests that closing one
 * PooledConnection does not delete the temporary destinations of another
 * PooledConnection that uses the same underlying ConnectionPool.
 */
@Timeout(60)
public class PooledConnectionTempDestCleanupTest extends ActiveMQJmsPoolTestSupport {

    protected Connection directConn1;
    protected Connection directConn2;

    protected JmsPoolConnectionFactory pooledConnFact;
    protected Connection pooledConn1;
    protected Connection pooledConn2;

    protected TemporaryQueue tempDest;
    protected TemporaryQueue otherTempDest;

    /**
     * Prepare to run a test case: create, configure, and start the embedded
     * broker, as well as creating the client connections to the broker.
     */
    @Override
    @BeforeEach
    public void setUp(TestInfo info) throws Exception {
        super.setUp(info);

        // Create the JmsPoolConnectionFactory.
        pooledConnFact = createPooledConnectionFactory();

        // Prepare the connections
        directConn1 = amqFactory.createConnection();
        directConn1.start();
        directConn2 = amqFactory.createConnection();
        directConn2.start();

        pooledConn1 = pooledConnFact.createConnection();
        pooledConn1.start();
        pooledConn2 = pooledConnFact.createConnection();
        pooledConn2.start();
    }

    @Override
    @AfterEach
    public void tearDown() throws java.lang.Exception {
        try {
            pooledConn1.stop();
        } catch (JMSException jms_exc) {
        }
        try {
            pooledConn2.stop();
        } catch (JMSException jms_exc) {
        }
        try {
            directConn1.close();
        } catch (JMSException jms_exc) {
        }
        try {
            directConn2.close();
        } catch (JMSException jms_exc) {
        }

        super.tearDown();
    }

    /*
     * Test for lingering temporary destinations after closing a
     * PooledConnection. Here are the steps:
     *
     * 1. create a session on the first pooled connection 2. create a session on
     * the second pooled connection 3. create a temporary destination on the
     * first session 4. confirm the temporary destination exists in the broker
     * 5. close the first connection 6. check that the temporary destination no
     * longer exists in the broker
     */
    @Test
    public void testPooledLingeringTempDests() throws java.lang.Exception {
        Session session1;
        Session session2;

        session1 = pooledConn1.createSession(false, Session.AUTO_ACKNOWLEDGE);
        session2 = pooledConn2.createSession(false, Session.AUTO_ACKNOWLEDGE);

        tempDest = session1.createTemporaryQueue();

        assertTrue(destinationExists(tempDest), "The expected temporary destination does not exist: " + tempDest);

        pooledConn1.close();

        assertTrue(!destinationExists(tempDest), "Temporary destination from closed pooled connection is lingering: " + tempDest);

        session2.close();
    }

    @Test
    public void testPooledTempDestsWithConsumer() throws java.lang.Exception {
        Session session1;

        session1 = pooledConn1.createSession(false, Session.AUTO_ACKNOWLEDGE);

        tempDest = session1.createTemporaryQueue();

        MessageConsumer messageConsumer = session1.createConsumer(tempDest);
        messageConsumer.receiveNoWait();

        assertTrue(destinationExists(tempDest), "TEST METHOD FAILURE - NEW TEMP DESTINATION DOES NOT EXIST");

        pooledConn1.close();

        assertTrue(!destinationExists(tempDest), "FAILED: temp dest from closed pooled connection is lingering");
    }

    /*
     * Test that closing one PooledConnection does not delete the temporary
     * destinations of another.
     *
     * 1. create a session on the first pooled connection 2. create a session on
     * the second pooled connection 3. create a temporary destination on the
     * first session 4. create a temporary destination on the second session 5.
     * confirm both temporary destinations exist in the broker 6. close the
     * first connection 7. check that the first temporary destination no longer
     * exists in the broker 8. check that the second temporary destination does
     * still exist in the broker
     */
    @Test
    public void testPooledTempDestsCleanupOverzealous() throws java.lang.Exception {
        Session session1;
        Session session2;

        session1 = pooledConn1.createSession(false, Session.AUTO_ACKNOWLEDGE);
        session2 = pooledConn2.createSession(false, Session.AUTO_ACKNOWLEDGE);

        tempDest = session1.createTemporaryQueue();
        otherTempDest = session2.createTemporaryQueue();

        assertTrue(destinationExists(tempDest), "The expected temporary destination does not exist: " + tempDest);
        assertTrue(destinationExists(otherTempDest), "The expected temporary destination does not exist: " + otherTempDest);

        pooledConn1.close();

        // Now confirm the first temporary destination no longer exists and the
        // second does.
        assertTrue(!destinationExists(tempDest), "Temporary destination from closed pooled connection is lingering: " + tempDest);
        assertTrue(destinationExists(otherTempDest), "Second PooledConnection's temporary destination was incorrectly deleted");
    }

    /*
     * CONTROL CASE
     *
     * Test for lingering temporary destinations after closing a Connection that
     * is NOT pooled. This demonstrates the standard JMS operation and helps to
     * validate the test methodology.
     *
     * 1. create a session on the first direct connection 2. create a session on
     * the second direct connection 3. create a temporary destination on the
     * first session 4. confirm the destination exists in the broker 5. close
     * the first connection 6. check that the destination no longer exists in
     * the broker
     */
    @Test
    public void testDirectLingeringTempDests() throws java.lang.Exception {
        Session session1;
        Session session2;

        session1 = directConn1.createSession(false, Session.AUTO_ACKNOWLEDGE);
        session2 = directConn2.createSession(false, Session.AUTO_ACKNOWLEDGE);

        tempDest = session1.createTemporaryQueue();

        assertTrue(destinationExists(tempDest), "The expected temporary destination does not exist: " + tempDest);

        directConn1.close();

        // Now confirm the temporary destination no longer exists.
        assertTrue(!destinationExists(tempDest), "Temporary destination from closed pooled connection is lingering: " + tempDest);

        session2.close();
    }

    @Test
    public void testDirectLingeringTempDestsWithConsumer() throws java.lang.Exception {
        Session session1;

        session1 = directConn1.createSession(false, Session.AUTO_ACKNOWLEDGE);

        tempDest = session1.createTemporaryQueue();

        MessageConsumer messageConsumer = session1.createConsumer(tempDest);
        messageConsumer.receiveNoWait();

        assertTrue(destinationExists(tempDest), "TEST METHOD FAILURE - NEW TEMP DESTINATION DOES NOT EXIST");

        directConn1.close();

        // Now confirm the temporary destination no longer exists.
        assertTrue((!destinationExists(tempDest)), "CONTROL TEST FAILURE - TEST METHOD IS SUSPECT");
    }

    private boolean destinationExists(Destination dest) throws Exception {
        RegionBroker rb = (RegionBroker) brokerService.getBroker().getAdaptor(RegionBroker.class);
        return rb.getTopicRegion().getDestinationMap().containsKey(dest) || rb.getQueueRegion().getDestinationMap().containsKey(dest)
                || rb.getTempTopicRegion().getDestinationMap().containsKey(dest) || rb.getTempQueueRegion().getDestinationMap().containsKey(dest);
    }
}
